from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from rpe.engines import Engine
from rpe.resources import Resource


@dataclass
class Policy:
    policy_id: str
    engine: Engine
    applies_to: List[str]
    description: str = ""


@dataclass
class Evaluation:
    resource: Resource
    engine: Engine
    policy_id: str

    compliant: bool
    remediable: bool

    # Static attributes on a given policy
    # Examples: severity, description, owner, etc.
    policy_attributes: Optional[Dict[str, Any]] = field(default_factory=dict)

    # Calculated attributes when evaluating the policy against a resource
    # Examples:
    #   * Why was a resource (non)compliant?
    #   * The prior `excluded` attribute can be implemented here
    evaluation_attributes: Optional[Dict[str, Any]] = field(default_factory=dict)

    def remediate(self):
        return self.engine.remediate(self.resource, self.policy_id)

    # This is to support the same interface as previous versions
    # and may be removed in a future release
    @property
    def excluded(self) -> bool:
        return self.evaluation_attributes.get("excluded", False)



# Copyright 2020 The resource-policy-evaluation-library Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections

import pytest
from google.oauth2.credentials import Credentials

from rpe.exceptions import ResourceException
from rpe.resources.gcp import (
    GcpAppEngineInstance,
    GcpBigqueryDataset,
    GcpBigtableInstance,
    GcpCloudFunction,
    GcpComputeDisk,
    GcpComputeFirewall,
    GcpComputeInstance,
    GcpComputeRegionDisk,
    GcpComputeSubnetwork,
    GcpDataflowJob,
    GcpDatafusionInstance,
    GcpDataprocCluster,
    GcpGkeCluster,
    GcpGkeClusterNodepool,
    GcpIamServiceAccount,
    GcpIamServiceAccountKey,
    GcpMemcacheInstance,
    GcpOrganization,
    GcpProject,
    GcpProjectService,
    GcpPubsubSubscription,
    GcpPubsubTopic,
    GcpRedisInstance,
    GcpSqlInstance,
    GcpStorageBucket,
    GoogleAPIResource,
)

client_kwargs = {"credentials": Credentials(token="")}

CaiTestCase = collections.namedtuple("CaiTestCase", "data resource_cls")

test_cases = [
    CaiTestCase(
        data={
            "name": "//appengine.googleapis.com/apps/test-project/services/default/versions/1234567890/instances/test-instance",
            "asset_type": "appengine.googleapis.com/Instance",
        },
        resource_cls=GcpAppEngineInstance,
    ),
    CaiTestCase(
        data={
            "name": "//bigquery.googleapis.com/projects/test-project/datasets/test-resource",
            "asset_type": "bigquery.googleapis.com/Dataset",
        },
        resource_cls=GcpBigqueryDataset,
    ),
    CaiTestCase(
        data={
            "name": "//bigtable.googleapis.com/projects/test-project/instances/test-resource",
            "asset_type": "bigtableadmin.googleapis.com/Instance",
        },
        resource_cls=GcpBigtableInstance,
    ),
    CaiTestCase(
        data={
            "name": "//cloudfunctions.googleapis.com/projects/test-project/locations/us-central1/functions/my-function",
            "asset_type": "cloudfunctions.googleapis.com/CloudFunction",
        },
        resource_cls=GcpCloudFunction,
    ),
    CaiTestCase(
        data={
            "name": "//compute.googleapis.com/projects/test-project/zones/us-central1-a/instances/test-resource",
            "asset_type": "compute.googleapis.com/Instance",
        },
        resource_cls=GcpComputeInstance,
    ),
    CaiTestCase(
        data={
            "name": "//compute.googleapis.com/projects/test-project/zones/us-central1-a/disks/test-resource",
            "asset_type": "compute.googleapis.com/Disk",
        },
        resource_cls=GcpComputeDisk,
    ),
    CaiTestCase(
        data={
            "name": "//compute.googleapis.com/projects/test-project/regions/us-central1/disks/test-resource",
            "asset_type": "compute.googleapis.com/RegionDisk",
        },
        resource_cls=GcpComputeRegionDisk,
    ),
    CaiTestCase(
        data={
            "name": "//compute.googleapis.com/projects/test-project/regions/asia-east2/subnetworks/test-resource",
            "asset_type": "compute.googleapis.com/Subnetwork",
        },
        resource_cls=GcpComputeSubnetwork,
    ),
    CaiTestCase(
        data={
            "name": "//compute.googleapis.com/projects/test-project/global/firewalls/test-resource",
            "asset_type": "compute.googleapis.com/Firewall",
        },
        resource_cls=GcpComputeFirewall,
    ),
    CaiTestCase(
        data={
            "name": "//datafusion.googleapis.com/projects/test-project/locations/us-central1/instances/test-resource",
            "asset_type": "datafusion.googleapis.com/Instance",
        },
        resource_cls=GcpDatafusionInstance,
    ),
    CaiTestCase(
        data={
            "name": "//dataproc.googleapis.com/projects/test-project/regions/us-central1/clusters/test-resource",
            "asset_type": "dataproc.googleapis.com/Cluster",
        },
        resource_cls=GcpDataprocCluster,
    ),
    CaiTestCase(
        data={
            "name": "//container.googleapis.com/projects/test-project/locations/us-central1/clusters/test-resource",
            "asset_type": "container.googleapis.com/Cluster",
        },
        resource_cls=GcpGkeCluster,
    ),
    CaiTestCase(
        data={
            "name": "//container.googleapis.com/projects/test-project/locations/us-central1/clusters/foofoo123/nodePools/test-resource",
            "asset_type": "container.googleapis.com/NodePool",
        },
        resource_cls=GcpGkeClusterNodepool,
    ),
    CaiTestCase(
        data={
            "name": "//iam.googleapis.com/projects/test-project/serviceAccounts/foo",
            "asset_type": "iam.googleapis.com/ServiceAccount",
        },
        resource_cls=GcpIamServiceAccount,
    ),
    CaiTestCase(
        data={
            "name": "//iam.googleapis.com/projects/test-project/serviceAccounts/foo/keys/bar",
            "asset_type": "iam.googleapis.com/ServiceAccountKey",
        },
        resource_cls=GcpIamServiceAccountKey,
    ),
    CaiTestCase(
        data={
            "name": "//pubsub.googleapis.com/projects/test-project/subscriptions/test-resource",
            "asset_type": "pubsub.googleapis.com/Subscription",
        },
        resource_cls=GcpPubsubSubscription,
    ),
    CaiTestCase(
        data={
            "name": "//pubsub.googleapis.com/projects/test-project/topics/test-resource",
            "asset_type": "pubsub.googleapis.com/Topic",
        },
        resource_cls=GcpPubsubTopic,
    ),
    CaiTestCase(
        data={
            "name": "//storage.googleapis.com/test-resource",
            "asset_type": "storage.googleapis.com/Bucket",
        },
        resource_cls=GcpStorageBucket,
    ),
    CaiTestCase(
        data={
            "name": "//cloudsql.googleapis.com/projects/test-project/instances/test-resource",
            "asset_type": "sqladmin.googleapis.com/Instance",
        },
        resource_cls=GcpSqlInstance,
    ),
    CaiTestCase(
        data={
            "name": "//cloudresourcemanager.googleapis.com/organizations/test-resource",
            "asset_type": "cloudresourcemanager.googleapis.com/Organization",
        },
        resource_cls=GcpOrganization,
    ),
    CaiTestCase(
        data={
            "name": "//cloudresourcemanager.googleapis.com/projects/test-resource",
            "asset_type": "cloudresourcemanager.googleapis.com/Project",
        },
        resource_cls=GcpProject,
    ),
    CaiTestCase(
        data={
            "name": "//serviceusage.googleapis.com/projects/179590471157/services/test-resource",
            "asset_type": "serviceusage.googleapis.com/Service",
        },
        resource_cls=GcpProjectService,
    ),
    CaiTestCase(
        data={
            "name": "//dataflow.googleapis.com/projects/test-project/locations/us-east1/jobs/test-resource",
            "asset_type": "dataflow.googleapis.com/Job",
        },
        resource_cls=GcpDataflowJob,
    ),
    CaiTestCase(
        data={
            "name": "//redis.googleapis.com/projects/test-project/locations/us-central1/instances/test-resource",
            "asset_type": "redis.googleapis.com/Instance",
        },
        resource_cls=GcpRedisInstance,
    ),
    CaiTestCase(
        data={
            "name": "//memcache.googleapis.com/projects/test-project/locations/us-central1/instances/test-resource",
            "asset_type": "memcache.googleapis.com/Instance",
        },
        resource_cls=GcpMemcacheInstance,
    ),
]


@pytest.mark.parametrize(
    "case", test_cases, ids=[case.resource_cls.__name__ for case in test_cases]
)
def test_gcp_resource_from_cai_data(case):
    r = GoogleAPIResource.from_cai_data(
        case.data.get("name"),
        case.data.get("asset_type"),
        client_kwargs=client_kwargs,
    )
    assert r.__class__ == case.resource_cls
    assert r.full_resource_name() == case.data.get("name")


def test_bad_resource_type():

    with pytest.raises(ResourceException) as excinfo:
        GoogleAPIResource.from_cai_data(
            "//cloudfakeservice.googleapis.com/widgets/test-resource",
            "cloudfakeservice.googleapis.com/Widget",
            client_kwargs=client_kwargs,
        )

    assert "Unrecognized resource type" in str(excinfo.value)



from rpe.resources.gcp import GoogleAPIResource
from googleapiclient.http import HttpMockSequence
import pytest


@pytest.mark.parametrize(
    "name,asset_type,http,inferred_key",
    [
        pytest.param(
            "//storage.googleapis.com/buckets/test-bucket",
            "storage.googleapis.com/Bucket",
            HttpMockSequence(
                [
                    ({"status": 200}, '{"location": "US"}'),
                    ({"status": 200}, '{"iam_policy":""}'),
                ]
            ),
            "location",
            id="bucket_location",
        ),
        pytest.param(
            "//compute.googleapis.com/projects/test-project/zones/test-zone/instances/test-instance",
            "compute.googleapis.com/Instance",
            HttpMockSequence(
                [
                    ({"status": 200}, '{"id": "1234567890"}'),
                    ({"status": 200}, '{"iam_policy":""}'),
                ]
            ),
            "uniquifier",
            id="instance_uniquifier",
        ),
    ],
)
def test_inferred_data_lookup_success(name, asset_type, http, inferred_key):
    res = GoogleAPIResource.from_cai_data(name, asset_type, http=http)

    assert inferred_key not in res._resource_data
    res.to_dict()
    assert inferred_key in res._resource_data


@pytest.mark.parametrize(
    "name,asset_type,http",
    [
        pytest.param(
            "//storage.googleapis.com/buckets/test-bucket",
            "storage.googleapis.com/Bucket",
            HttpMockSequence(
                [
                    ({"status": 401}, '{"msg": "oops"}'),
                ]
            ),
            id="resource_401",
        ),
        pytest.param(
            "//compute.googleapis.com/projects/test-project/zones/test-zone/instances/test-instance",
            "compute.googleapis.com/Instance",
            HttpMockSequence(
                [
                    ({"status": 404}, '{"msg": "something went wrong"}'),
                ]
            ),
            id="resource_404",
        ),
        pytest.param(
            "//compute.googleapis.com/projects/test-project/zones/test-zone/instances/test-instance",
            "compute.googleapis.com/Instance",
            HttpMockSequence(
                [
                    ({"status": 200}, "HTTP ERROR: NOT JSON"),
                ]
            ),
            id="malformed_json",
        ),
    ],
)
def test_todict_eats_exceptions(name, asset_type, http):
    res = GoogleAPIResource.from_cai_data(name, asset_type, http=http)

    res.to_dict()



# Copyright 2021 The resource-policy-evaluation-library Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from datetime import datetime
from typing import Optional
from pydantic import BaseModel

from google.pubsub_v1 import PubsubMessage
import jmespath


class PubsubMessageMetadata(BaseModel):
    publish_time: datetime
    message_id: str
    attributes: Optional[dict]


def get_pubsub_message_metadata(message: PubsubMessage):
    return PubsubMessageMetadata(
        publish_time=message.publish_time,
        message_id=message.message_id,
        attributes=dict(message.attributes),
    )



# Copyright 2019 The resource-policy-evaluation-library Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
from urllib.parse import urlparse

import jmespath
import tenacity
from googleapiclienthelpers.discovery import build_subresource
from googleapiclienthelpers.waiter import Waiter

from rpe.exceptions import (
    InvalidRemediationSpecStep,
    ResourceException,
    UnsupportedRemediationSpec,
    is_retryable_exception,
)

from .base import Resource


class GoogleAPIResource(Resource):

    # Names of the get method of the root resource
    get_method = "get"
    required_resource_data = ["name"]

    # jmespath expression for getting labels
    resource_labels_path = "resource.labels"

    # Other properties of a resource we might need to perform evaluations, such as iam policy
    resource_components = {}

    # If a resource is not in a ready state, we can't update it. If we retrieve
    # it, and the state changes, updates will be rejected because the ETAG will
    # have changed. If a resource defines readiness criteria, the get() call
    # will wait until the resource is in a ready state to return
    #
    # Key/Value to check to see if a resource is ready
    readiness_key = None
    readiness_value = None
    readiness_terminal_values = []

    inferred_data_map = None

    def __init__(self, client_kwargs=None, http=None, **resource_data):

        if client_kwargs is None:
            client_kwargs = {}

        self._http = http
        # Set some defaults
        self._service = None
        self._resource_metadata = None
        self._full_resource_name = None

        # Load and validate additional resource data
        self._resource_data = resource_data
        self._validate_resource_data()

        # Store the client kwargs to pass to any new clients
        self._client_kwargs = client_kwargs

        self._ancestry = None

    def _validate_resource_data(self):
        """Verify we have all the required data for this resource"""
        if not all(arg in self._resource_data for arg in self.required_resource_data):

            raise ResourceException(
                "Missing data required for resource creation. Expected data: {}; Got: {}".format(
                    ",".join(self.required_resource_data),
                    ",".join(self._resource_data.keys()),
                )
            )

    @staticmethod
    def _extract_cai_name_data(name):
        """Attempt to get identifiable information out of a Cloud Asset Inventory-formatted resource_name"""

        # Most resources need only a subset of these fields to query the google apis
        fields = {
            "project_id": r"/projects/([^\/]+)/",
            "location": r"/(?:locations|regions|zones)/([^\/]+)/",
            "name": r"([^\/]+)$",
            # Less-common resource data
            #  AppEngine
            "app": r"/apps/([^\/]+)/",
            "service": r"/services/([^\/]+)/",
            "version": r"/versions/([^\/]+)/",
            #  NodePools
            "cluster": r"/clusters/([^\/]+)/",
            # ServiceAccounts
            "service_account": r"serviceAccounts/([^\/]+)/",
        }

        resource_data = {}

        # Extract available resource data from resource name
        for field_name in fields:
            m = re.search(fields[field_name], name)
            if m:
                resource_data[field_name] = m.group(1)

        return resource_data

    @classmethod
    def subclass_by_type(cls, resource_type):
        mapper = {res_cls.resource_type: res_cls for res_cls in cls.__subclasses__()}

        try:
            return mapper[resource_type]
        except KeyError:
            raise ResourceException(
                "Unrecognized resource type: {}".format(resource_type)
            )

    @classmethod
    def from_resource_data(
        cls, *, resource_type, client_kwargs=None, http=None, **resource_data
    ):
        if client_kwargs is None:
            client_kwargs = {}
        res_cls = cls.subclass_by_type(resource_type)
        return res_cls(client_kwargs=client_kwargs, http=http, **resource_data)

    @staticmethod
    def from_cai_data(
        resource_name, resource_type, project_id=None, client_kwargs=None, http=None
    ):
        """Attempt to return the appropriate resource using Cloud Asset Inventory-formatted resource info"""

        if client_kwargs is None:
            client_kwargs = {}

        res_cls = GoogleAPIResource.subclass_by_type(resource_type)

        resource_data = GoogleAPIResource._extract_cai_name_data(resource_name)

        # if the project_id was passed, and its wasnt found in the resource name, add it
        if project_id and "project_id" not in resource_data:
            resource_data["project_id"] = project_id

        return res_cls(client_kwargs=client_kwargs, http=http, **resource_data)

    def to_dict(self):

        self._refresh_inferred_data()

        details = self._resource_data.copy()
        details.update(
            {
                "resource_type": self.resource_type,
            }
        )

        try:
            details["full_resource_name"] = self.full_resource_name()
        except Exception:
            details["full_resource_name"] = None

        return details

    # Some useful resource data may not be available when instantiated
    def _refresh_inferred_data(self):

        if not self.inferred_data_map:
            return

        try:
            resource_metadata = self.get(refresh=False)["resource"]
        except Exception:
            return

        for key, path in self.inferred_data_map.items():
            # Don't replace existing data
            if key in self._resource_data:
                continue
            value = jmespath.search(path, resource_metadata)

            if value is not None:
                self._resource_data[key] = value

    def type(self):
        return self.resource_type

    def full_resource_name(self):
        if self._full_resource_name is None:
            self.gen_full_resource_name()

        return self._full_resource_name

    # Google's documentation describes what it calls a 'full resource name' for
    # resources. None of the API's seem to implement it (except Cloud Asset
    # Inventory). This attempts to generate it from the discovery-based api
    # client's generated http request url.
    #
    # If we inject it into the resource, we can use it in policy evaluation to
    # simplify the structure of our policies
    def gen_full_resource_name(self):

        method = getattr(self.service, self.get_method)
        uri = method(**self._get_request_args()).uri

        uri_parsed = urlparse(uri)
        domain = uri_parsed.netloc
        path_segments = uri_parsed.path[1:].split("/")

        # CAI uses cloudsql.googleapis.com in their full_resource_name, so we need to detect all
        # bad incarnations and replace them
        bad_sql_names = ["sql", "sqladmin"]

        # First we need the name of the api
        if domain.startswith("www."):
            # we need to get the api name from the path
            api_name = path_segments.pop(0)
            api_name = "cloudsql" if api_name in bad_sql_names else api_name
        else:
            # the api name is the first segment of the domain
            api_name = domain.split(".")[0]

            # the sql api is now returning sqladmin.googleapis.com/sql/<ver>
            # and the CAI docs state the FRN for sql instances should start
            # with //cloudsql.googleapis.com/ so lets replace all odd sql ones
            # and rely on the code below to catch duplicates
            path_segments[0] = (
                "cloudsql" if path_segments[0] in bad_sql_names else path_segments[0]
            )
            api_name = "cloudsql" if api_name in bad_sql_names else api_name

            # occasionally the compute api baseUrl is returned as
            # compute.googleapis.com/compute, in which case we need to remove
            # the duplicated api reference
            # also addresses the sql issue mentioned above
            if api_name == path_segments[0]:
                path_segments.pop(0)

        # Remove the version from the path
        path_segments.pop(0)

        # Remove method from the last path segment
        if ":" in path_segments[-1]:
            path_segments[-1] = path_segments[-1].split(":")[0]

        # Annoying resource-specific fixes

        # The url for buckets is `/b/` instead of `/buckets/`. Initially this fixed that
        # Unfortunately, CAI omits the collection name, rather than follow Google's API design doc
        # So we just remove the collection name
        #
        # https://cloud.google.com/apis/design/resource_names
        # See: https://issuetracker.google.com/issues/131586763
        #
        if api_name == "storage" and path_segments[0] == "b":
            path_segments.pop(0)
            # Replace with this if they fix CAI:
            # path_segments[0] = "buckets"

        if api_name == "bigtableadmin":
            api_name = "bigtable"

        resource_path = "/".join(path_segments)

        self._full_resource_name = "//{}.googleapis.com/{}".format(
            api_name, resource_path
        )

    def _get_component(self, component):
        method_name = self.resource_components[component]

        # Many components take the same request signature, but allow for custom request
        # args if needed. Fall back to default args if the expected function doesn't exist
        if hasattr(self, f"_get_{component}_request_args"):
            req_arg_method = getattr(self, f"_get_{component}_request_args")
        else:
            req_arg_method = getattr(self, "_get_request_args")

        method = getattr(self.service, method_name)

        component_metadata = method(**req_arg_method()).execute()
        return component_metadata

    def get(self, refresh=True):

        if not refresh and self._resource_metadata:
            return self._resource_metadata

        method = getattr(self.service, self.get_method)

        # If the resource has readiness criteria, wait for it
        if self.readiness_key and self.readiness_value:
            waiter = Waiter(method, **self._get_request_args())
            asset = waiter.wait(
                self.readiness_key,
                self.readiness_value,
                terminal_values=self.readiness_terminal_values,
                interval=10,
                retries=90,
            )
        else:
            asset = method(**self._get_request_args()).execute()

        resp = {
            "type": self.type(),
            "name": self.full_resource_name(),
        }

        resp["resource"] = asset

        for c in self.resource_components:
            resp[c] = self._get_component(c)

        self._resource_metadata = resp
        return self._resource_metadata

    # Determine what remediation steps to take, allow for future remediation specifications
    def remediate(self, remediation):
        # Check for an update spec version, default to version 1
        remediation_spec = remediation.get("_remediation_spec", "")

        if remediation_spec in ["v2beta1", "v2"]:
            required_keys = ["method", "params"]

            for step in remediation.get("steps", []):
                if not all(k in step for k in required_keys):
                    raise InvalidRemediationSpecStep()

                method_name = step.get("method")
                params = step.get("params")
                self._call_method(method_name, params)
        else:
            raise UnsupportedRemediationSpec(
                "The specified remediation spec is not supported"
            )

    @tenacity.retry(
        retry=tenacity.retry_if_exception(is_retryable_exception),
        wait=tenacity.wait_random_exponential(multiplier=5, max=20),
        stop=tenacity.stop_after_attempt(15),
    )
    def _call_method(self, method_name, params):
        """Call the requested method on the resource"""
        method = getattr(self.service, method_name)
        return method(**params).execute()

    @property
    def ancestry(self):
        if self._ancestry:
            return self._ancestry

        # attempt to fill in the resource's ancestry
        # if the target project has the cloudresourcemanager api disabled, this will fail
        # if the resource_data doesn't include the project_id (ex: with storage buckets) this will also fail
        try:
            resource_manager_projects = build_subresource(
                "cloudresourcemanager.projects",
                "v1",
                **self._client_kwargs,
                http=self._http,
            )

            resp = resource_manager_projects.getAncestry(
                projectId=self.project_id
            ).execute()

            # Reformat getAncestry response to be a list of resource names
            self._ancestry = [
                f"//cloudresourcemanager.googleapis.com/{ancestor['resourceId']['type']}s/{ancestor['resourceId']['id']}"
                for ancestor in resp.get("ancestor")
            ]

        except Exception:
            # This call is best-effort. Any failures should be caught
            pass

        return self._ancestry

    @property
    def organization(self):
        ancestry = self.ancestry
        if not ancestry:
            return None

        return next(
            (
                ancestor
                for ancestor in ancestry
                if ancestor.startswith(
                    "//cloudresourcemanager.googleapis.com/organizations/"
                )
            ),
            None,
        )

    @property
    def client_kwargs(self):
        return self._client_kwargs

    @client_kwargs.setter
    def client_kwargs(self, client_kwargs):

        # Invalidate service/parent because client_kwargs changed
        self._service = None

        self._client_kwargs = client_kwargs

    @property
    def service(self):
        if self._service is None:

            full_resource_path = "{}.{}".format(self.service_name, self.resource_path)
            self._service = build_subresource(
                full_resource_path, self.version, **self._client_kwargs, http=self._http
            )
        return self._service

    @property
    def labels(self):
        labels = (
            jmespath.search(self.resource_labels_path, self.get(refresh=False)) or {}
        )
        if not isinstance(labels, dict):
            raise TypeError("Unexpected label format")

        return labels

    @property
    def resource_name(self):
        return self._resource_data.get("name")

    @property
    def project_id(self):
        return self._resource_data.get("project_id")

    @property
    def location(self):
        return self._resource_data.get("location")

    # The full_resource_name is universally unique, but only at a point in time. If a resource is deleted,
    # and a new one created, with the same name/location/etc, the full_resource_name alone won't distinguish
    # those 2 resources. This field attempts to define an identifier that, when combined with the
    # full_resource_name, should be universally unique over time. In some cases, google actually
    # provides resource-specific unique IDs (for example: the compute API). In other cases we may use the
    # creation timestamp of the resource. Not all resources have fields that make this  possible.
    @property
    def uniquifier(self):
        self._refresh_inferred_data()

        return self._resource_data.get("uniquifier")


class GcpAppEngineInstance(GoogleAPIResource):

    service_name = "appengine"
    resource_path = "apps.services.versions.instances"
    version = "v1"
    readiness_key = "vmStatus"
    readiness_value = "RUNNING"

    resource_type = "appengine.googleapis.com/Instance"  # this is made-up based on existing appengine types

    required_resource_data = ["name", "app", "service", "version"]

    inferred_data_map = {
        "uniquifier": "startTime",
    }

    def _get_request_args(self):
        return {
            "appsId": self._resource_data["app"],
            "servicesId": self._resource_data["service"],
            "versionsId": self._resource_data["version"],
            "instancesId": self._resource_data["name"],
        }


class GcpBigqueryDataset(GoogleAPIResource):

    service_name = "bigquery"
    resource_path = "datasets"
    version = "v2"

    required_resource_data = ["name", "project_id"]

    resource_type = "bigquery.googleapis.com/Dataset"

    inferred_data_map = {
        "uniquifier": "id",
    }

    def _get_request_args(self):
        return {
            "datasetId": self._resource_data["name"],
            "projectId": self._resource_data["project_id"],
        }


class GcpBigtableInstance(GoogleAPIResource):

    service_name = "bigtableadmin"
    resource_path = "projects.instances"
    version = "v2"
    readiness_key = "state"
    readiness_value = "READY"

    resource_components = {
        "iam": "getIamPolicy",
    }

    required_resource_data = ["name", "project_id"]

    resource_type = "bigtableadmin.googleapis.com/Instance"

    def _get_request_args(self):
        return {
            "name": "projects/{}/instances/{}".format(
                self._resource_data["project_id"], self._resource_data["name"]
            ),
        }

    def _get_iam_request_args(self):
        return {
            "resource": "projects/{}/instances/{}".format(
                self._resource_data["project_id"], self._resource_data["name"]
            ),
        }


class GcpCloudFunction(GoogleAPIResource):

    service_name = "cloudfunctions"
    resource_path = "projects.locations.functions"
    version = "v1"

    resource_components = {
        "iam": "getIamPolicy",
    }
    readiness_key = "status"
    readiness_value = "ACTIVE"
    readiness_terminal_values = [
        "CLOUD_FUNCTION_STATUS_UNSPECIFIED",
        "OFFLINE",
        "UNKNOWN",
    ]

    required_resource_data = ["name", "location", "project_id"]

    resource_type = "cloudfunctions.googleapis.com/CloudFunction"  # unreleased

    def _get_request_args(self):
        return {
            "name": "projects/{}/locations/{}/functions/{}".format(
                self._resource_data["project_id"],
                self._resource_data["location"],
                self._resource_data["name"],
            ),
        }

    # The top-level dict key is different
    def _get_iam_request_args(self):
        return {
            "resource": "projects/{}/locations/{}/functions/{}".format(
                self._resource_data["project_id"],
                self._resource_data["location"],
                self._resource_data["name"],
            ),
        }


class GcpComputeInstance(GoogleAPIResource):

    service_name = "compute"
    resource_path = "instances"
    version = "v1"

    required_resource_data = ["name", "location", "project_id"]

    resource_type = "compute.googleapis.com/Instance"

    inferred_data_map = {
        "uniquifier": "id",
    }

    def _get_request_args(self):
        return {
            "instance": self._resource_data["name"],
            "zone": self._resource_data["location"],
            "project": self._resource_data["project_id"],
        }


class GcpComputeDisk(GoogleAPIResource):

    service_name = "compute"
    resource_path = "disks"
    version = "v1"

    required_resource_data = ["name", "location", "project_id"]

    resource_type = "compute.googleapis.com/Disk"

    inferred_data_map = {
        "uniquifier": "id",
    }

    def _get_request_args(self):
        return {
            "project": self._resource_data["project_id"],
            "zone": self._resource_data["location"],
            "disk": self._resource_data["name"],
        }


class GcpComputeRegionDisk(GoogleAPIResource):

    service_name = "compute"
    resource_path = "regionDisks"
    version = "v1"

    required_resource_data = ["name", "location", "project_id"]

    resource_type = "compute.googleapis.com/RegionDisk"

    inferred_data_map = {
        "uniquifier": "id",
    }

    def _get_request_args(self):
        return {
            "project": self._resource_data["project_id"],
            "region": self._resource_data["location"],
            "disk": self._resource_data["name"],
        }


class GcpComputeNetwork(GoogleAPIResource):

    service_name = "compute"
    resource_path = "networks"
    version = "v1"

    required_resource_data = ["name", "project_id"]

    resource_type = "compute.googleapis.com/Network"

    inferred_data_map = {
        "uniquifier": "id",
    }

    def _get_request_args(self):
        return {
            "project": self._resource_data["project_id"],
            "network": self._resource_data["name"],
        }


class GcpComputeSubnetwork(GoogleAPIResource):

    service_name = "compute"
    resource_path = "subnetworks"
    version = "v1"

    required_resource_data = ["name", "location", "project_id"]

    resource_type = "compute.googleapis.com/Subnetwork"

    inferred_data_map = {
        "uniquifier": "id",
    }

    def _get_request_args(self):
        return {
            "project": self._resource_data["project_id"],
            "region": self._resource_data["location"],
            "subnetwork": self._resource_data["name"],
        }


class GcpComputeFirewall(GoogleAPIResource):

    service_name = "compute"
    resource_path = "firewalls"
    version = "v1"

    required_resource_data = ["name", "project_id"]

    resource_type = "compute.googleapis.com/Firewall"

    inferred_data_map = {
        "uniquifier": "id",
    }

    def _get_request_args(self):
        return {
            "firewall": self._resource_data["name"],
            "project": self._resource_data["project_id"],
        }


class GcpDataprocCluster(GoogleAPIResource):
    service_name = "dataproc"
    resource_path = "projects.regions.clusters"
    version = "v1beta2"

    required_resource_data = ["name", "location", "project_id"]

    resource_type = "dataproc.googleapis.com/Cluster"

    inferred_data_map = {
        "uniquifier": "clusterUuid",
    }

    def _get_request_args(self):
        return {
            "projectId": self._resource_data["project_id"],
            "region": self._resource_data["location"],
            "clusterName": self._resource_data["name"],
        }


class GcpDatafusionInstance(GoogleAPIResource):
    service_name = "datafusion"
    resource_path = "projects.locations.instances"
    version = "v1beta1"

    required_resource_data = ["name", "location", "project_id"]

    readiness_key = "state"
    readiness_value = "RUNNING"
    readiness_terminal_values = ["FAILED", "DELETING", "STATUS_UNSPECIFIED"]

    resource_components = {
        "iam": "getIamPolicy",
    }

    resource_type = "datafusion.googleapis.com/Instance"

    inferred_data_map = {
        "uniquifier": "createTime",
    }

    def _get_request_args(self):
        return {
            "name": "projects/{}/locations/{}/instances/{}".format(
                self.project_id,
                self._resource_data["location"],
                self._resource_data["name"],
            )
        }

    def _get_iam_request_args(self):
        return {
            "resource": "projects/{}/locations/{}/instances/{}".format(
                self.project_id,
                self._resource_data["location"],
                self._resource_data["name"],
            )
        }


class GcpGkeCluster(GoogleAPIResource):

    service_name = "container"
    resource_path = "projects.locations.clusters"
    version = "v1"
    readiness_key = "status"
    readiness_value = "RUNNING"
    readiness_terminal_values = ["ERROR", "DEGRADED", "STOPPING", "STATUS_UNSPECIFIED"]

    required_resource_data = ["name", "location", "project_id"]

    resource_labels_path = "resource.resourceLabels"

    resource_type = "container.googleapis.com/Cluster"

    inferred_data_map = {
        "uniquifier": "id",
    }

    def _get_request_args(self):
        return {
            "name": "projects/{}/locations/{}/clusters/{}".format(
                self._resource_data["project_id"],
                self._resource_data["location"],
                self._resource_data["name"],
            )
        }


class GcpGkeClusterNodepool(GoogleAPIResource):

    service_name = "container"
    resource_path = "projects.locations.clusters.nodePools"
    version = "v1"
    readiness_key = "status"
    readiness_value = "RUNNING"
    readiness_terminal_values = [
        "ERROR",
        "RUNNING_WITH_ERROR",
        "RECONCILING",
        "STOPPING",
    ]

    required_resource_data = ["name", "cluster", "location", "project_id"]

    resource_type = "container.googleapis.com/NodePool"  # beta

    def _get_request_args(self):
        return {
            "name": "projects/{}/locations/{}/clusters/{}/nodePools/{}".format(
                self._resource_data["project_id"],
                self._resource_data["location"],
                self._resource_data["cluster"],
                self._resource_data["name"],
            )
        }


class GcpIamServiceAccount(GoogleAPIResource):

    service_name = "iam"
    resource_path = "projects.serviceAccounts"
    version = "v1"

    required_resource_data = ["name", "project_id"]

    resource_type = "iam.googleapis.com/ServiceAccount"

    inferred_data_map = {
        "uniquifier": "uniqueId",
    }

    def _get_request_args(self):
        return {
            "name": "projects/{}/serviceAccounts/{}".format(
                self._resource_data["project_id"], self._resource_data["name"]
            )
        }


class GcpIamServiceAccountKey(GoogleAPIResource):

    service_name = "iam"
    resource_path = "projects.serviceAccounts.keys"
    version = "v1"

    required_resource_data = ["name", "service_account", "project_id"]

    resource_type = "iam.googleapis.com/ServiceAccountKey"

    inferred_data_map = {
        "uniquifier": "privateKeyData",
    }

    def _get_request_args(self):
        return {
            "name": "projects/{}/serviceAccounts/{}/keys/{}".format(
                self._resource_data["project_id"],
                self._resource_data["service_account"],
                self._resource_data["name"],
            )
        }


class GcpPubsubSubscription(GoogleAPIResource):

    service_name = "pubsub"
    resource_path = "projects.subscriptions"
    version = "v1"

    required_resource_data = ["name", "project_id"]

    resource_components = {
        "iam": "getIamPolicy",
    }

    resource_type = "pubsub.googleapis.com/Subscription"

    def _get_request_args(self):
        return {
            "subscription": "projects/{}/subscriptions/{}".format(
                self._resource_data["project_id"], self._resource_data["name"]
            )
        }

    def _get_iam_request_args(self):
        return {
            "resource": "projects/{}/subscriptions/{}".format(
                self._resource_data["project_id"], self._resource_data["name"]
            )
        }


class GcpPubsubTopic(GoogleAPIResource):

    service_name = "pubsub"
    resource_path = "projects.topics"
    version = "v1"

    required_resource_data = ["name", "project_id"]

    resource_components = {
        "iam": "getIamPolicy",
    }

    resource_type = "pubsub.googleapis.com/Topic"

    def _get_request_args(self):
        return {
            "topic": "projects/{}/topics/{}".format(
                self._resource_data["project_id"], self._resource_data["name"]
            )
        }

    def _get_iam_request_args(self):
        return {
            "resource": "projects/{}/topics/{}".format(
                self._resource_data["project_id"], self._resource_data["name"]
            )
        }


class GcpStorageBucket(GoogleAPIResource):

    service_name = "storage"
    resource_path = "buckets"
    version = "v1"

    resource_components = {
        "iam": "getIamPolicy",
    }

    required_resource_data = ["name"]

    resource_type = "storage.googleapis.com/Bucket"

    inferred_data_map = {
        "location": "location",
        "uniquifier": "timeCreated",
    }

    def _get_request_args(self):
        return {
            "bucket": self._resource_data["name"],
        }


class GcpSqlInstance(GoogleAPIResource):

    service_name = "sqladmin"
    resource_path = "instances"
    version = "v1beta4"
    readiness_key = "state"
    readiness_value = "RUNNABLE"
    readiness_terminal_values = ["FAILED", "MAINTENANCE", "SUSPENDED", "UNKNOWN_STATE"]

    resource_labels_path = "resource.settings.userLabels"

    resource_type = "sqladmin.googleapis.com/Instance"

    def _get_request_args(self):
        return {
            "instance": self._resource_data["name"],
            "project": self._resource_data["project_id"],
        }


class GcpOrganization(GoogleAPIResource):

    service_name = "cloudresourcemanager"
    resource_path = "organizations"
    version = "v1"

    resource_components = {
        "iam": "getIamPolicy",
    }

    resource_type = "cloudresourcemanager.googleapis.com/Organization"

    def _get_request_args(self):
        return {"name": "organizations/" + self._resource_data["name"]}

    def _get_iam_request_args(self):
        return {"resource": "organizations/" + self._resource_data["name"]}


class GcpProject(GoogleAPIResource):

    service_name = "cloudresourcemanager"
    resource_path = "projects"
    version = "v1"

    resource_components = {
        "iam": "getIamPolicy",
    }

    resource_type = "cloudresourcemanager.googleapis.com/Project"  # beta

    inferred_data_map = {
        "uniquifier": "projectNumber",
    }

    def _get_request_args(self):
        return {"projectId": self._resource_data["name"]}

    def _get_iam_request_args(self):
        return {"resource": self._resource_data["name"], "body": {}}


class GcpProjectService(GoogleAPIResource):

    service_name = "serviceusage"
    resource_path = "services"
    version = "v1"

    required_resource_data = ["name", "project_id"]

    resource_type = "serviceusage.googleapis.com/Service"

    def _get_request_args(self):
        return {
            "name": "projects/{}/services/{}".format(
                self._resource_data["project_id"], self._resource_data["name"]
            )
        }


class GcpDataflowJob(GoogleAPIResource):

    service_name = "dataflow"
    resource_path = "projects.locations.jobs"
    version = "v1b3"

    required_resource_data = ["name", "project_id", "location"]

    resource_type = "dataflow.googleapis.com/Job"

    inferred_data_map = {
        "uniquifier": "id",
    }

    def _get_request_args(self):
        return {
            "jobId": self._resource_data["name"],
            "projectId": self._resource_data["project_id"],
            "location": self._resource_data["location"],
            "view": "JOB_VIEW_DESCRIPTION",
        }


class GcpRedisInstance(GoogleAPIResource):

    service_name = "redis"
    resource_path = "projects.locations.instances"
    version = "v1"

    readiness_key = "state"
    readiness_value = "READY"
    readiness_terminal_values = ["DELETING", "STATE_UNSPECIFIED"]

    required_resource_data = ["name", "project_id", "location"]

    resource_type = "redis.googleapis.com/Instance"

    inferred_data_map = {
        "uniquifier": "createTime",
    }

    def _get_request_args(self):
        return {
            "name": "projects/{}/locations/{}/instances/{}".format(
                self._resource_data["project_id"],
                self._resource_data["location"],
                self._resource_data["name"],
            ),
        }


class GcpMemcacheInstance(GoogleAPIResource):

    service_name = "memcache"
    resource_path = "projects.locations.instances"
    version = "v1"

    readiness_key = "state"
    readiness_value = "READY"
    readiness_terminal_values = ["DELETING", "STATE_UNSPECIFIED"]

    required_resource_data = ["name", "project_id", "location"]

    resource_type = "memcache.googleapis.com/Instance"

    inferred_data_map = {
        "uniquifier": "createTime",
    }

    def _get_request_args(self):
        return {
            "name": "projects/{}/locations/{}/instances/{}".format(
                self._resource_data["project_id"],
                self._resource_data["location"],
                self._resource_data["name"],
            ),
        }



# Copyright 2019 The resource-policy-evaluation-library Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections

import pytest
from google.oauth2.credentials import Credentials
from googleapiclient.http import HttpMockSequence

from rpe.exceptions import ResourceException
from rpe.resources.gcp import (
    GcpAppEngineInstance,
    GcpBigqueryDataset,
    GcpBigtableInstance,
    GcpCloudFunction,
    GcpComputeDisk,
    GcpComputeFirewall,
    GcpComputeInstance,
    GcpComputeSubnetwork,
    GcpDataflowJob,
    GcpDatafusionInstance,
    GcpDataprocCluster,
    GcpGkeCluster,
    GcpGkeClusterNodepool,
    GcpMemcacheInstance,
    GcpOrganization,
    GcpProject,
    GcpProjectService,
    GcpPubsubSubscription,
    GcpPubsubTopic,
    GcpRedisInstance,
    GcpSqlInstance,
    GcpStorageBucket,
    GoogleAPIResource,
)

test_project = "my_project"
test_resource_name = "my_resource"
client_kwargs = {"credentials": Credentials(token="")}

ResourceTestCase = collections.namedtuple(
    "ResourceTestCase", "resource_data cls resource_type name http uniquifier"
)

test_cases = [
    ResourceTestCase(
        resource_data={
            "name": "my_resource",
            "app": "my_project",
            "service": "default",
            "version": "0123456789",
            "project_id": test_project,
        },
        cls=GcpAppEngineInstance,
        resource_type="appengine.googleapis.com/Instance",
        name="//appengine.googleapis.com/apps/my_project/services/default/versions/0123456789/instances/my_resource",
        http=HttpMockSequence(
            [
                (
                    {"status": 200},
                    '{"name":"name", "startTime":"appEngine_startTime", "vmStatus" : "RUNNING"}',
                ),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier="appEngine_startTime",
    ),
    ResourceTestCase(
        resource_data={"name": test_resource_name, "project_id": test_project},
        cls=GcpBigqueryDataset,
        resource_type="bigquery.googleapis.com/Dataset",
        name="//bigquery.googleapis.com/projects/my_project/datasets/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name", "id":"dataset_id"}'),
            ]
        ),
        uniquifier="dataset_id",
    ),
    ResourceTestCase(
        resource_data={"name": test_resource_name, "project_id": test_project},
        cls=GcpBigtableInstance,
        resource_type="bigtableadmin.googleapis.com/Instance",
        name="//bigtable.googleapis.com/projects/my_project/instances/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name", "state":"READY"}'),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier=None,
    ),
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "us-central1-a",
            "project_id": test_project,
        },
        cls=GcpCloudFunction,
        resource_type="cloudfunctions.googleapis.com/CloudFunction",
        name="//cloudfunctions.googleapis.com/projects/my_project/locations/us-central1-a/functions/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name", "status":"ACTIVE"}'),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier=None,
    ),
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "us-central1-a",
            "project_id": test_project,
        },
        cls=GcpComputeDisk,
        resource_type="compute.googleapis.com/Disk",
        name="//compute.googleapis.com/projects/my_project/zones/us-central1-a/disks/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name", "id":"disk_id"}'),
            ]
        ),
        uniquifier="disk_id",
    ),
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "us-central1-a",
            "project_id": test_project,
        },
        cls=GcpComputeInstance,
        resource_type="compute.googleapis.com/Instance",
        name="//compute.googleapis.com/projects/my_project/zones/us-central1-a/instances/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name", "id":"disk_id"}'),
            ]
        ),
        uniquifier="disk_id",
    ),
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "us-central1-a",
            "project_id": test_project,
        },
        cls=GcpGkeCluster,
        resource_type="container.googleapis.com/Cluster",
        name="//container.googleapis.com/projects/my_project/locations/us-central1-a/clusters/my_resource",
        http=HttpMockSequence(
            [
                (
                    {"status": 200},
                    '{"name":"name", "id":"cluster_id", "status" : "RUNNING"}',
                ),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier="cluster_id",
    ),
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "us-central1-a",
            "cluster": "parent_resource",
            "project_id": test_project,
        },
        cls=GcpGkeClusterNodepool,
        resource_type="container.googleapis.com/NodePool",
        name="//container.googleapis.com/projects/my_project/locations/us-central1-a/clusters/parent_resource/nodePools/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name", "status" : "RUNNING"}'),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier=None,
    ),
    ResourceTestCase(
        resource_data={
            "name": "012345678901",
        },
        cls=GcpOrganization,
        resource_type="cloudresourcemanager.googleapis.com/Organization",
        name="//cloudresourcemanager.googleapis.com/organizations/012345678901",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name"}'),
            ]
        ),
        uniquifier=None,
    ),
    ResourceTestCase(
        resource_data={"name": test_project, "project_id": test_project},
        cls=GcpProject,
        resource_type="cloudresourcemanager.googleapis.com/Project",
        name="//cloudresourcemanager.googleapis.com/projects/my_project",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name", "projectNumber":"project_number"}'),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier="project_number",
    ),
    ResourceTestCase(
        resource_data={"name": "compute.googleapis.com", "project_id": test_project},
        cls=GcpProjectService,
        resource_type="serviceusage.googleapis.com/Service",
        name="//serviceusage.googleapis.com/projects/my_project/services/compute.googleapis.com",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name"}'),
            ]
        ),
        uniquifier=None,
    ),
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "us-central1",
            "project_id": test_project,
        },
        cls=GcpDatafusionInstance,
        resource_type="datafusion.googleapis.com/Instance",
        name="//datafusion.googleapis.com/projects/my_project/locations/us-central1/instances/my_resource",
        http=HttpMockSequence(
            [
                (
                    {"status": 200},
                    '{"name":"name", "createTime":"datafusion_createTime" , "state":"RUNNING"}',
                ),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier="datafusion_createTime",
    ),
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "global",
            "project_id": test_project,
        },
        cls=GcpDataprocCluster,
        resource_type="dataproc.googleapis.com/Cluster",
        name="//dataproc.googleapis.com/projects/my_project/regions/global/clusters/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name", "clusterUuid":"cluster_Uuid"}'),
            ]
        ),
        uniquifier="cluster_Uuid",
    ),
    ResourceTestCase(
        resource_data={"name": test_resource_name, "project_id": test_project},
        cls=GcpPubsubSubscription,
        resource_type="pubsub.googleapis.com/Subscription",
        name="//pubsub.googleapis.com/projects/my_project/subscriptions/my_resource",
        http=HttpMockSequence([({"status": 200}, '{"name":"name"}')]),
        uniquifier=None,
    ),
    ResourceTestCase(
        resource_data={"name": test_resource_name, "project_id": test_project},
        cls=GcpPubsubTopic,
        resource_type="pubsub.googleapis.com/Topic",
        name="//pubsub.googleapis.com/projects/my_project/topics/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name"}'),
            ]
        ),
        uniquifier=None,
    ),
    ResourceTestCase(
        resource_data={"name": test_resource_name, "project_id": test_project},
        cls=GcpSqlInstance,
        resource_type="sqladmin.googleapis.com/Instance",
        name="//cloudsql.googleapis.com/projects/my_project/instances/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name" , "state":"RUNNABLE"}'),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier=None,
    ),
    ResourceTestCase(
        resource_data={"name": test_resource_name, "project_id": test_project},
        cls=GcpStorageBucket,
        resource_type="storage.googleapis.com/Bucket",
        # This should include the collection name `/buckets/`, but CAI doesn't do that
        # See: https://issuetracker.google.com/issues/131586763
        name="//storage.googleapis.com/my_resource",
        http=HttpMockSequence(
            [
                (
                    {"status": 200},
                    '{"name":"name", "timeCreated":"bucket_timeCreated"}',
                ),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier="bucket_timeCreated",
    ),
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "us-central1",
            "project_id": test_project,
        },
        cls=GcpComputeSubnetwork,
        resource_type="compute.googleapis.com/Subnetwork",
        name="//compute.googleapis.com/projects/my_project/regions/us-central1/subnetworks/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name", "id":"subnetwork_id"}'),
            ]
        ),
        uniquifier="subnetwork_id",
    ),
    ResourceTestCase(
        resource_data={"name": test_resource_name, "project_id": test_project},
        cls=GcpComputeFirewall,
        resource_type="compute.googleapis.com/Firewall",
        name="//compute.googleapis.com/projects/my_project/global/firewalls/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name", "id":"firewall_id"}'),
            ]
        ),
        uniquifier="firewall_id",
    ),
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "us-central1",
            "project_id": test_project,
        },
        cls=GcpDataflowJob,
        resource_type="dataflow.googleapis.com/Job",
        name="//dataflow.googleapis.com/projects/my_project/locations/us-central1/jobs/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"name":"name", "id":"dataflow_id"}'),
            ]
        ),
        uniquifier="dataflow_id",
    ),
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "us-central1",
            "project_id": test_project,
        },
        cls=GcpRedisInstance,
        resource_type="redis.googleapis.com/Instance",
        name="//redis.googleapis.com/projects/my_project/locations/us-central1/instances/my_resource",
        http=HttpMockSequence(
            [
                (
                    {"status": 200},
                    '{"name":"name", "createTime":"redis_createTime", "state":"READY"}',
                ),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier="redis_createTime",
    ),
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "us-central1",
            "project_id": test_project,
        },
        cls=GcpMemcacheInstance,
        resource_type="memcache.googleapis.com/Instance",
        name="//memcache.googleapis.com/projects/my_project/locations/us-central1/instances/my_resource",
        http=HttpMockSequence(
            [
                (
                    {"status": 200},
                    '{"name":"name", "createTime":"memcache_createTime", "state":"READY"}',
                ),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier="memcache_createTime",
    ),
]

location_test_cases = [
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "us-central1-a",
            "project_id": test_project,
        },
        cls=GcpRedisInstance,
        resource_type="redis.googleapis.com/Instance",
        name="//redis.googleapis.com/projects/my_project/locations/us-central1/instances/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"createTime":"createTime"}'),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier="startTime",
    ),
    ResourceTestCase(
        resource_data={
            "name": test_resource_name,
            "location": "us-central1",
            "project_id": test_project,
        },
        cls=GcpMemcacheInstance,
        resource_type="memcache.googleapis.com/Instance",
        name="//memcache.googleapis.com/projects/my_project/locations/us-central1/instances/my_resource",
        http=HttpMockSequence(
            [
                ({"status": 200}, '{"createTime":"createTime"}'),
                ({"status": 200}, "{}"),
            ]
        ),
        uniquifier="createTime",
    ),
]


@pytest.mark.parametrize(
    "case", test_cases, ids=[case.cls.__name__ for case in test_cases]
)
def test_gcp_from_resource(case):
    r = GoogleAPIResource.from_resource_data(
        resource_type=case.resource_type, http=case.http, **case.resource_data
    )
    assert r.__class__ == case.cls
    assert isinstance(r._get_request_args(), dict)
    assert r.uniquifier == case.uniquifier


def test_gcp_from_resource_no_type():
    with pytest.raises(TypeError) as excinfo:
        GoogleAPIResource.from_resource_data(project_id=test_project)

    assert "required keyword-only argument" in str(excinfo.value)
    assert "resource_type" in str(excinfo.value)


def test_gcp_resource_bad_type():
    with pytest.raises(ResourceException) as excinfo:
        GoogleAPIResource.from_resource_data(resource_type="fake.type")

    assert "Unrecognized resource type" in str(excinfo.value)


@pytest.mark.parametrize(
    "case", test_cases, ids=[case.cls.__name__ for case in test_cases]
)
def test_gcp_full_resource_name(case):
    r = GoogleAPIResource.from_resource_data(
        resource_type=case.resource_type,
        client_kwargs=client_kwargs,
        **case.resource_data
    )
    assert r.full_resource_name() == case.name


@pytest.mark.parametrize(
    "case", location_test_cases, ids=[case.cls.__name__ for case in location_test_cases]
)
def test_gcp_location(case):
    r = GoogleAPIResource.from_resource_data(
        resource_type=case.resource_type,
        client_kwargs=client_kwargs,
        **case.resource_data
    )
    assert r.location == case.resource_data.get("location")


def test_missing_resource_data():
    with pytest.raises(ResourceException) as excinfo:

        GcpAppEngineInstance(name=test_resource_name)

    assert "Missing data required for resource creation" in str(excinfo.value)


def test_gcp_to_dict():
    r = GoogleAPIResource.from_resource_data(
        resource_type="storage.googleapis.com/Bucket",
        client_kwargs=client_kwargs,
        name=test_resource_name,
        project_id=test_project,
    )

    data = r.to_dict()
    # with no creds, we should still get this key but it should be none
    assert data["project_id"] == test_project



from abc import ABC, abstractmethod


# Base class for policy engines
class Engine(ABC):
    @abstractmethod
    def policies(self):
        pass

    @abstractmethod
    def evaluate(self, resource):
        pass

    @abstractmethod
    def remediate(self, resource, policy_id):
        pass



# Copyright 2019 The resource-policy-evaluation-library Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod


class Resource(ABC):

    # Returns a dictionary representing the resource. Must contain a 'type' key
    # indicating what type of resource it is
    @abstractmethod
    def get(self):
        pass

    # Performs remediation based on a json representation of how to remediate a
    # resource that does not comply to a given policy. This allows for
    # remediation from non-python based engines, such as the open-policy-agent
    # engine
    @abstractmethod
    def remediate(self):
        pass

    # Returns the resource type
    @abstractmethod
    def type(self):
        pass



# Copyright 2019 The resource-policy-evaluation-library Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import importlib.util
import inspect
import sys

from rpe.policy import Evaluation, Policy


class PythonPolicyEngine:

    counter = 0

    def __init__(self, package_path):

        self._policies = {}
        self.package_path = package_path
        PythonPolicyEngine.counter += 1
        self.package_name = "rpe.plugins.policies.py_" + str(PythonPolicyEngine.counter)

        self._load_policies()

    def _load_policies(self):
        spec = importlib.util.spec_from_file_location(
            self.package_name, "{}/__init__.py".format(self.package_path)
        )
        module = importlib.util.module_from_spec(spec)
        sys.modules[self.package_name] = module

        spec.loader.exec_module(module)

        for name, obj in inspect.getmembers(module):
            if (
                inspect.isclass(obj)
                and hasattr(obj, "applies_to")
                and isinstance(obj.applies_to, list)
            ):
                self._policies[name] = obj

    def policies(self):
        """
        Returns:
            A list of names of configured policies
        """

        policies = [
            Policy(
                policy_id=policy_name,
                engine=self,
                applies_to=policy_cls.applies_to,
                description=policy_cls.description,
            )
            for policy_name, policy_cls in self._policies.items()
        ]

        return policies

    def evaluate(self, resource):
        matched_policies = dict(
            filter(
                lambda policy: resource.type() in policy[1].applies_to,
                self._policies.items(),
            )
        )

        # Loop over policy and build evals, so we can catch exceptions

        evals = []

        for policy_name, policy_cls in matched_policies.items():
            try:

                if hasattr(policy_cls, "evaluate"):
                    evals.append(policy_cls.evaluate(self, resource))

                else:
                    evals.append(self._legacy_eval(resource, policy_name, policy_cls))

            # These are user-provided modules, we need to catch any exception
            except Exception as e:
                print(f"Evaluation exception. Policy: {policy_name}, Message: {str(e)}")

        return evals

    def _legacy_eval(self, resource, policy_name, policy_cls):
        compliant = policy_cls.compliant(resource) is True

        eval_attr = {"excluded": policy_cls.excluded(resource) is True}

        ev = Evaluation(
            resource=resource,
            engine=self,
            policy_id=policy_name,
            compliant=compliant,
            remediable=hasattr(policy_cls, "remediate"),
            evaluation_attributes=eval_attr,
        )

        return ev

    def remediate(self, resource, policy_id):
        policy_cls = self._policies[policy_id]
        policy_cls.remediate(resource)



# Copyright 2021 The resource-policy-evaluation-library Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from dataclasses import asdict, dataclass
from typing import Optional
from pydantic import BaseModel

from rpe.extractors import Extractor
from rpe.extractors.models import ExtractedMetadata, ExtractedResources
from rpe.extractors.pubsub_metadata import (
    PubsubMessageMetadata,
    get_pubsub_message_metadata,
)
from rpe.resources.gcp import GoogleAPIResource


class MicromanagerMetadata(PubsubMessageMetadata, ExtractedMetadata):
    pass


class MicromanagerEvaluationRequest(Extractor):
    @classmethod
    def extract(cls, message):

        message_data = json.loads(message.data)

        name = message_data.get("name")
        asset_type = message_data.get("asset_type")
        project_id = message_data.get("project_id")
        metadata = message_data.get("metadata") or {}
        metadata.update((get_pubsub_message_metadata(message)).dict())

        resource = GoogleAPIResource.from_cai_data(
            name, asset_type, project_id=project_id
        )

        metadata = MicromanagerMetadata(**metadata)

        return ExtractedResources(resources=[resource], metadata=metadata)



# Copyright 2019 The resource-policy-evaluation-library Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from setuptools import find_packages, setup

setup(
    name="rpe-lib",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="A resource policy evaluation library",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Joe Ceresini",
    url="https://github.com/cleardataeng/resource-policy-evaluation-library",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    install_requires=[
        "google-api-python-client",
        "google-api-python-client-helpers",
        "jmespath",
        "tenacity",
        "python-dateutil",
    ],
    packages=find_packages(include=["rpe*"]),
    package_data={},
    license="Apache 2.0",
    keywords="gcp policy enforcement",
)
